/*_############################################################################
  _## 
  _##  xmibform.cpp  
  _##
  _##  SNMP++v3.2.10
  _##  -----------------------------------------------
  _##  Copyright (c) 2001-2004 Jochen Katz, Frank Fock
  _##
  _##  This software is based on SNMP++2.6 from Hewlett Packard:
  _##  
  _##    Copyright (c) 1996
  _##    Hewlett-Packard Company
  _##  
  _##  ATTENTION: USE OF THIS SOFTWARE IS SUBJECT TO THE FOLLOWING TERMS.
  _##  Permission to use, copy, modify, distribute and/or sell this software 
  _##  and/or its documentation is hereby granted without fee. User agrees 
  _##  to display the above copyright notice and this license notice in all 
  _##  copies of the software and any documentation of the software. User 
  _##  agrees to assume all liability for the use of the software; 
  _##  Hewlett-Packard and Jochen Katz make no representations about the 
  _##  suitability of this software for any purpose. It is provided 
  _##  "AS-IS" without warranty of any kind, either express or implied. User 
  _##  hereby grants a royalty-free license to any and all derivatives based
  _##  upon this software code base. 
  _##  
  _##  Stuttgart, Germany, Tue Jan 13 01:13:13 CET 2004 
  _##  
  _##########################################################################*/
/*

  Copyright (c) 1996
  Hewlett-Packard Company

  ATTENTION: USE OF THIS SOFTWARE IS SUBJECT TO THE FOLLOWING TERMS.
  Permission to use, copy, modify, distribute and/or sell this software
  and/or its documentation is hereby granted without fee. User agrees
  to display the above copyright notice and this license notice in all
  copies of the software and any documentation of the software. User
  agrees to assume all liability for the use of the software; Hewlett-Packard
  makes no representations about the suitability of this software for any
  purpose. It is provided "AS-IS without warranty of any kind,either express
  or implied. User hereby grants a royalty-free license to any and all
  derivatives based upon this software code base.

*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <Xm/Xm.h>
#include <Xm/MwmUtil.h>
#include <X11/Shell.h>
#include <Xm/MenuShell.h>

#include <Xm/PushB.h>
#include <Xm/Text.h>
#include <Xm/ScrolledW.h>
#include <Xm/MessageB.h>
#include <Xm/TextF.h>
#include <Xm/Label.h>
#include <Xm/Form.h>

#include <snmp_pp.h>

#include "xmibform.h"

extern Widget		UxTopLevel;
extern Snmp		*the_Session;
extern Widget		FormHandle;

static Widget           dialogWgt = NULL;

void ErrorDialog(char *errorString){
  Widget tmpWgt;
  Arg args[5];
  int i=0;
  XmString message;
  XmString title;

  message = XmStringCreateSimple(errorString);
  title = XmStringCreateSimple("Error");

    /*---------------------------------------------------------------*/
    /* Create a standard Motif message box, but hide the Cancel and  */
    /* Help buttons, because the user just gets to look at the error */
    /* no other action is available...                               */
    /*---------------------------------------------------------------*/
  if (dialogWgt == NULL){
    XtSetArg(args[i], XmNdialogStyle, XmDIALOG_APPLICATION_MODAL); i++;
    XtSetArg(args[i], XmNdialogTitle, title); i++;

    dialogWgt = XmCreateErrorDialog(FormHandle, "errorDialog", args, i);
    tmpWgt = XmMessageBoxGetChild(dialogWgt, XmDIALOG_CANCEL_BUTTON);
    XtUnmanageChild(tmpWgt);
    tmpWgt = XmMessageBoxGetChild(dialogWgt, XmDIALOG_HELP_BUTTON);
    XtUnmanageChild(tmpWgt);
  }

  XtVaSetValues(dialogWgt, 
		XmNmessageString, message, 
		NULL);

    /*-----------------------------------------*/
    /* Toss the strings... Can we do this yet? */
    /*-----------------------------------------*/
  XmStringFree(message);
  XmStringFree(title);

    /*------------------------*/
    /* Now display the window */
    /*------------------------*/
  XtManageChild(dialogWgt);
}

void GetNextCB( int reason, Snmp *, Pdu &pdu, SnmpTarget &, void *cd){
  Vb nextVb;
  char buf[700];
  CQueryForm *displayForm = (CQueryForm *) cd;
  int pdu_error;

  displayForm->enableGetNext();

  if (reason != SNMP_CLASS_ASYNC_RESPONSE){
    sprintf(buf, "Error returned by GetNext:\n %s", the_Session->error_msg(reason));
    ErrorDialog(buf);
    displayForm->displayText(buf);
    displayForm->displayText("\n\n");
    return;
  }

  pdu_error = pdu.get_error_status();
  if (pdu_error){
    sprintf(buf, "Response contains error:\n %s", the_Session->error_msg(pdu_error));
    ErrorDialog(buf);
    displayForm->displayText(buf);
    displayForm->displayText("\n\n");
    return;
  }

  pdu.get_vb(nextVb, 0);
  sprintf(buf, "  %s:  %s\n\n", 
	  nextVb.get_printable_oid(), 
	  nextVb.get_printable_value());
  displayForm->displayText(buf);
  displayForm->updateOid(nextVb.get_printable_oid());
}

void IssueGetNext(char *hostname, char* port, char *oid, char *commun, snmp_version version, 
		  CQueryForm *displayForm){
  UdpAddress dest(hostname);

  u_short iport = atoi(port);
  if (iport==0)
	iport = 161;

  dest.set_port(iport);

  int status;

  if (!dest.valid()){
    char buf[500];
    sprintf(buf, "Couldn't resolve hostname for: %s", hostname);
    ErrorDialog(buf);
    displayForm->displayText(buf);
    displayForm->displayText("\n\n");
    displayForm->enableGetNext();
    return;
  }
  OctetStr community(commun);
  
  CTarget target(dest);
  Vb vb = Vb(oid);
  Pdu pdu;
  target.set_version( version );
  target.set_timeout( 500 );
  target.set_retry( 2 );
  target.set_readcommunity( community);
  pdu += vb;
  status = the_Session->get_next(pdu, target, GetNextCB, (void *) displayForm);

  if (status){
    char buf[500];
    sprintf(buf, "Error issuing GetNext:\n %s", the_Session->error_msg(status));
    ErrorDialog(buf);
    displayForm->displayText(buf);
    displayForm->displayText("\n\n");
    displayForm->enableGetNext();
    return;
  }
};

void IssueGetNextv3(char *hostname, char *port, char *oid, 
                    snmp_version version, char *securityName,
                    int securityLevel, char *contextName, 
                    char *contextEngineID,  CQueryForm *displayForm)
{
  UdpAddress dest(hostname);
      
  u_short iport = atoi(port);
  if (iport==0)
    iport = 161;
  
  dest.set_port(iport);
  
  int status;
  
  if (!dest.valid()){
    char buf[500];
    sprintf(buf, "Couldn't resolve hostname for: %s", hostname);
    ErrorDialog(buf);
    displayForm->displayText(buf);
    displayForm->displayText("\n\n");
    displayForm->enableGetNext();
    return;
  }

  int secModel = SecurityModel_USM;
  
  UTarget target(dest);
  Vb vb3= Vb (oid);
  Pdu pdu;
  target.set_version( version );
  target.set_timeout( 500 );
  target.set_retry( 2 );
  target.set_security_name(securityName);
  pdu.set_security_level(securityLevel);
  target.set_security_model(secModel);
  pdu.set_context_engine_id(contextEngineID);
  pdu.set_context_name(contextName);

  pdu += vb3;
  status = the_Session->get_next(pdu, target, GetNextCB, (void *) displayForm);
  if (status){
    char buf[500];
    sprintf(buf, "Error issuing GetNext:\n %s", the_Session->error_msg(status));
    ErrorDialog(buf);
    displayForm->displayText(buf);
    displayForm->displayText("\n\n");
    displayForm->enableGetNext();
    return;
  }
}


/*******************************************************************************
       The following are callback functions.
*******************************************************************************/

void  CQueryForm :: activateCB_pbQuery( Widget /*wgt*/, XtPointer, XtPointer)
{
    char *hostname;
    char *oidname;
    char *port;
    char *community;
    char *securityName;
    char *tmpSecurityLevel;
    char *contextName;
    char *contextEngineID;

    int securityLevel;

    snmp_version version = version3;
    char buf[500]; // FIXME
    char *tmp;
    int tmpint;

    hostname = XmTextFieldGetString(tfHostname);
    oidname = XmTextFieldGetString(tfOid);
    port = XmTextFieldGetString(tfPort);
    community = XmTextFieldGetString(tfCommunity);
    securityName = XmTextFieldGetString(tfSecurityName);
    tmpSecurityLevel = XmTextFieldGetString(tfSecurityLevel);
    securityLevel = atoi(tmpSecurityLevel);
    contextName = XmTextFieldGetString(tfContextName);
    contextEngineID = XmTextFieldGetString(tfContextEngineID);

    tmp = XmTextFieldGetString(tfVersion);
    tmpint = atoi(tmp);
    switch (tmpint) {
      case 1: {version = version1; break;}
      case 2: {version = version2c; break;}
      case 3: {version = version3; break;}
      default:{version = version1;}
    }
    if (version==version3) {
    sprintf(buf, "Issuing v3 GetNext(%s) to %s \nwith securityName %s, port %s...\n", 
	    oidname, hostname, securityName, port);
    displayText(buf);

    disableGetNext();
    IssueGetNextv3(hostname, port, oidname, version, securityName,
                   securityLevel, contextName, contextEngineID, this);
    } else {

    sprintf(buf, "Issuing v1/2 GetNext(%s) to %s \nwith community %s, port %s...\n", 
	    oidname, hostname, community, port);
    displayText(buf);

    disableGetNext();
    IssueGetNext(hostname, port, oidname, community, version, this);
    }
}

void  CQueryForm :: Wrap_activateCB_pbQuery( Widget wgt, 
				        XtPointer cd, 
					XtPointer cb)
{
	CQueryForm               *UxContext;
	UxContext = (CQueryForm *) cd;

	UxContext->activateCB_pbQuery(wgt, cd, cb);
}

void  CQueryForm :: activateCB_pbQuit( Widget, XtPointer, XtPointer)
{
   exit(0);
}

void  CQueryForm :: Wrap_activateCB_pbQuit(
					Widget wgt, 
					XtPointer cd, 
					XtPointer cb)
{
	CQueryForm               *UxContext;
	UxContext = (CQueryForm *) cd;

	UxContext->activateCB_pbQuit(wgt, cd, cb);
}

/*******************************************************************************
       The 'buildWidgets' function creates all the widgets for the
       main screen.
*******************************************************************************/

Widget CQueryForm::buildWidgets()
{
	Widget		theShell;

	// Creation of form1
	theShell = XtVaCreatePopupShell( "form1_shell",
		applicationShellWidgetClass, UxTopLevel,
		XmNx, 140,
		XmNy, 370,
		XmNwidth, 800,
		XmNheight, 700,
		XmNshellUnitType, XmPIXELS,
		XmNtitle, "SNMP Query",
		NULL );


	form1 = XtVaCreateManagedWidget( "form1",
			xmFormWidgetClass,
			theShell,
			XmNwidth, 800,
			XmNheight, 700,
			XmNresizePolicy, XmRESIZE_NONE,
			XmNunitType, XmPIXELS,
			NULL );


	// Creation of lHost
	lHost = XtVaCreateManagedWidget( "lHost",
			xmLabelWidgetClass,
			form1,
			XmNx, 10,
			XmNy, 10,
			XmNwidth, 110,
			XmNheight, 40,
			RES_CONVERT( XmNlabelString, "Hostname" ),
			NULL );


	// Creation of tfHostname
	tfHostname = XtVaCreateManagedWidget( "tfHostname",
			xmTextFieldWidgetClass,
			form1,
			XmNwidth, 290,
			XmNx, 130,
			XmNy, 10,
			XmNheight, 40,
			XmNleftAttachment, XmATTACH_WIDGET,
			XmNleftWidget, lHost,
			XmNrightAttachment, XmATTACH_FORM,
			XmNrightOffset, 10,
			NULL );

        //char hostnamebuf[100];
	//gethostname(hostnamebuf, 100);
        //XmTextFieldSetString(tfHostname, hostnamebuf);
        XmTextFieldSetString(tfHostname, "walapai.telematik.informatik.uni-karlsruhe.de");

        //****************************

	// Creation of lcommunity
	lCommunity = XtVaCreateManagedWidget( "lCommunity",
			xmLabelWidgetClass,
			form1,
			XmNx, 10,
			XmNy, 70,
			XmNwidth, 110,
			XmNheight, 40,
			RES_CONVERT( XmNlabelString, "Community" ),
			NULL );


	// Creation of tfcommunity
	tfCommunity = XtVaCreateManagedWidget( "tfCommunity",
			xmTextFieldWidgetClass,
			form1,
			XmNwidth, 290,
			XmNx, 130,
			XmNy, 70,
			XmNheight, 40,
			XmNleftAttachment, XmATTACH_WIDGET,
			XmNleftWidget, lCommunity,
			XmNrightAttachment, XmATTACH_FORM,
			XmNrightOffset, 10,
			NULL );

        XmTextFieldSetString(tfCommunity, "public");


        //***************************


	// Creation of lPort
	lPort = XtVaCreateManagedWidget( "lPort",
			xmLabelWidgetClass,
			form1,
			XmNx, 10,
			XmNy, 130,
			XmNwidth, 110,
			XmNheight, 40,
			RES_CONVERT( XmNlabelString, "Port" ),
			NULL );


	// Creation of tfPort
	tfPort = XtVaCreateManagedWidget( "tfPort",
			xmTextFieldWidgetClass,
			form1,
			XmNwidth, 50,
			XmNx, 120,
			XmNy, 130,
			XmNheight, 40,
			XmNleftAttachment, XmATTACH_WIDGET,
			XmNleftWidget, lPort,
			XmNrightAttachment, XmATTACH_FORM,
			XmNrightOffset, 10,
			NULL );

        XmTextFieldSetString(tfPort, "5001");

	// Creation of lOid
	lOid = XtVaCreateManagedWidget( "lOid",
			xmLabelWidgetClass,
			form1,
			XmNx, 10,
			XmNy, 180,
			XmNwidth, 110,
			XmNheight, 40,
			RES_CONVERT( XmNlabelString, "Object Id:" ),
			NULL );

	// Creation of tfOid
	tfOid = XtVaCreateManagedWidget( "tfOid",
			xmTextFieldWidgetClass,
			form1,
			XmNwidth, 430,
			XmNx, 120,
			XmNy, 180,
			XmNheight, 40,
			XmNleftAttachment, XmATTACH_WIDGET,
			XmNleftWidget, lOid,
			XmNrightAttachment, XmATTACH_FORM,
			XmNrightOffset, 10,
			NULL );

        XmTextFieldSetString(tfOid, "1.1");

	// Creation of lVersion
	lVersion = XtVaCreateManagedWidget( "lVersion",
			xmLabelWidgetClass,
			form1,
			XmNx, 10,
			XmNy, 230,
			XmNwidth, 110,
			XmNheight, 40,
			RES_CONVERT( XmNlabelString, "Version:" ),
			NULL );

	// Creation of tfVersion
	tfVersion = XtVaCreateManagedWidget( "tfVersion",
			xmTextFieldWidgetClass,
			form1,
			XmNwidth, 50,
			XmNx, 0,
			XmNy, 230,
			XmNheight, 40,
			XmNleftAttachment, XmATTACH_WIDGET,
			XmNleftWidget, lVersion,
			XmNrightAttachment, XmATTACH_FORM,
			XmNrightOffset, 10,
			NULL );

        XmTextFieldSetString(tfVersion, "3");

	// Creation of lSecurityName
	lSecurityName = XtVaCreateManagedWidget( "lSecurityNAme",
			xmLabelWidgetClass,
			form1,
			XmNx, 10,
			XmNy, 280,
			XmNwidth, 110,
			XmNheight, 40,
			RES_CONVERT( XmNlabelString, "securityName:" ),
			NULL );

	// Creation of tfSecurityName
	tfSecurityName = XtVaCreateManagedWidget( "tfSecurityName",
			xmTextFieldWidgetClass,
			form1,
			XmNwidth, 430,
			XmNx, 120,
			XmNy, 280,
			XmNheight, 40,
			XmNleftAttachment, XmATTACH_WIDGET,
			XmNleftWidget, lSecurityName,
			XmNrightAttachment, XmATTACH_FORM,
			XmNrightOffset, 10,
			NULL );

        XmTextFieldSetString(tfSecurityName, "SHAIDEA");

	// Creation of lSecurityLevel
	lSecurityLevel = XtVaCreateManagedWidget( "lSecurityLevel",
			xmLabelWidgetClass,
			form1,
			XmNx, 10,
			XmNy, 330,
			XmNwidth, 110,
			XmNheight, 40,
			RES_CONVERT( XmNlabelString, "SecurityLevel:" ),
			NULL );

	// Creation of tfSecurityLevel
	tfSecurityLevel = XtVaCreateManagedWidget( "tfSecurityLevel",
			xmTextFieldWidgetClass,
			form1,
			XmNwidth, 430,
			XmNx, 120,
			XmNy, 330,
			XmNheight, 40,
			XmNleftAttachment, XmATTACH_WIDGET,
			XmNleftWidget, lSecurityLevel,
			XmNrightAttachment, XmATTACH_FORM,
			XmNrightOffset, 10,
			NULL );

        XmTextFieldSetString(tfSecurityLevel, "3");

	// Creation of lContextEngineID
	lContextEngineID = XtVaCreateManagedWidget( "lContextEngineID",
			xmLabelWidgetClass,
			form1,
			XmNx, 10,
			XmNy, 380,
			XmNwidth, 110,
			XmNheight, 40,
			RES_CONVERT( XmNlabelString, "ContextEngineID:" ),
			NULL );

	// Creation of tfContextEngineID
	tfContextEngineID = XtVaCreateManagedWidget( "tfContextEngineID",
			xmTextFieldWidgetClass,
			form1,
			XmNwidth, 430,
			XmNx, 120,
			XmNy, 380,
			XmNheight, 40,
			XmNleftAttachment, XmATTACH_WIDGET,
			XmNleftWidget, lContextEngineID,
			XmNrightAttachment, XmATTACH_FORM,
			XmNrightOffset, 10,
			NULL );

        XmTextFieldSetString(tfContextEngineID, "Agent");

	// Creation of lContextName
	lContextName = XtVaCreateManagedWidget( "lContextName",
			xmLabelWidgetClass,
			form1,
			XmNx, 10,
			XmNy, 430,
			XmNwidth, 110,
			XmNheight, 40,
			RES_CONVERT( XmNlabelString, "ContextName:" ),
			NULL );

	// Creation of tfContextName
	tfContextName = XtVaCreateManagedWidget( "tfContextName",
			xmTextFieldWidgetClass,
			form1,
			XmNwidth, 430,
			XmNx, 120,
			XmNy, 430,
			XmNheight, 40,
			XmNleftAttachment, XmATTACH_WIDGET,
			XmNleftWidget, lContextName,
			XmNrightAttachment, XmATTACH_FORM,
			XmNrightOffset, 10,
			NULL );

        XmTextFieldSetString(tfContextName, "");

	// Creation of scrolledWindowText
	scrolledWindowText = XtVaCreateManagedWidget( "scrolledWindowText",
			xmScrolledWindowWidgetClass,
			form1,
			XmNscrollingPolicy, XmAPPLICATION_DEFINED,
			XmNvisualPolicy, XmVARIABLE,
			XmNscrollBarDisplayPolicy, XmSTATIC,
			XmNx, 30,
			XmNy, 4800,
			XmNbottomAttachment, XmATTACH_FORM,
			XmNbottomOffset, 60,
			XmNrightAttachment, XmATTACH_FORM,
			XmNrightOffset, 10,
			XmNleftAttachment, XmATTACH_FORM,
			XmNleftOffset, 20,
			XmNtopAttachment, XmATTACH_WIDGET,
			XmNtopOffset, 20,
			XmNtopWidget, lContextName,
			NULL );


	// Creation of scrolledText
	scrolledText = XtVaCreateManagedWidget( "scrolledText",
			xmTextWidgetClass,
			scrolledWindowText,
			XmNwidth, 500,
			XmNheight, 40,
			XmNeditMode, XmMULTI_LINE_EDIT ,
			XmNeditable, FALSE,
			NULL );


	// Creation of pbQuery
	pbQuery = XtVaCreateManagedWidget( "pbQuery",
			xmPushButtonWidgetClass,
			form1,
			XmNx, 20,
			XmNy, 630,
			XmNwidth, 150,
			XmNheight, 40,
			RES_CONVERT( XmNlabelString, "Get Next" ),
			XmNsensitive, TRUE,
			XmNbottomAttachment, XmATTACH_FORM,
			XmNleftOffset, 10,
			XmNtopAttachment, XmATTACH_NONE,
			XmNtopOffset, 0,
			XmNbottomOffset, 10,
			NULL );
	XtAddCallback( pbQuery, XmNactivateCallback,
		(XtCallbackProc) &CQueryForm::Wrap_activateCB_pbQuery,
		(XtPointer) this );


	// Creation of pbQuit
	pbQuit = XtVaCreateManagedWidget( "pbQuit",
			xmPushButtonWidgetClass,
			form1,
			XmNx, 390,
			XmNy, 630,
			XmNwidth, 150,
			XmNheight, 40,
			RES_CONVERT( XmNlabelString, "Quit" ),
			XmNbottomAttachment, XmATTACH_FORM,
			XmNbottomOffset, 10,
			XmNleftAttachment, XmATTACH_NONE,
			XmNleftOffset, 0,
			XmNrightAttachment, XmATTACH_FORM,
			XmNrightOffset, 10,
			XmNtopAttachment, XmATTACH_NONE,
			XmNtopOffset, 0,
			NULL );
	XtAddCallback( pbQuit, XmNactivateCallback,
		(XtCallbackProc) &CQueryForm::Wrap_activateCB_pbQuit,
		(XtPointer) this );


	return ( theShell );
}

/*******************************************************************************
       The following is the constructor function.
*******************************************************************************/

CQueryForm::CQueryForm( Widget UxParent )
{
	this->UxParent = UxParent;
}

/*******************************************************************************
       The following is the procedural creation function.  This is used
       instead of the constructor.  No real good reason for this...
*******************************************************************************/

Widget create_form1( Widget UxParent )
{
	CQueryForm *theInterface = 
			new CQueryForm( UxParent );
	return (theInterface->buildWidgets());
}
