/*_############################################################################
  _## 
  _##  gauge.h  
  _##
  _##  SNMP++v3.2.10
  _##  -----------------------------------------------
  _##  Copyright (c) 2001-2004 Jochen Katz, Frank Fock
  _##
  _##  This software is based on SNMP++2.6 from Hewlett Packard:
  _##  
  _##    Copyright (c) 1996
  _##    Hewlett-Packard Company
  _##  
  _##  ATTENTION: USE OF THIS SOFTWARE IS SUBJECT TO THE FOLLOWING TERMS.
  _##  Permission to use, copy, modify, distribute and/or sell this software 
  _##  and/or its documentation is hereby granted without fee. User agrees 
  _##  to display the above copyright notice and this license notice in all 
  _##  copies of the software and any documentation of the software. User 
  _##  agrees to assume all liability for the use of the software; 
  _##  Hewlett-Packard and Jochen Katz make no representations about the 
  _##  suitability of this software for any purpose. It is provided 
  _##  "AS-IS" without warranty of any kind, either express or implied. User 
  _##  hereby grants a royalty-free license to any and all derivatives based
  _##  upon this software code base. 
  _##  
  _##  Stuttgart, Germany, Tue Jan 13 01:13:13 CET 2004 
  _##  
  _##########################################################################*/
/*===================================================================

  Copyright (c) 1999
  Hewlett-Packard Company

  ATTENTION: USE OF THIS SOFTWARE IS SUBJECT TO THE FOLLOWING TERMS.
  Permission to use, copy, modify, distribute and/or sell this software
  and/or its documentation is hereby granted without fee. User agrees
  to display the above copyright notice and this license notice in all
  copies of the software and any documentation of the software. User
  agrees to assume all liability for the use of the software; Hewlett-Packard
  makes no representations about the suitability of this software for any
  purpose. It is provided "AS-IS without warranty of any kind,either express
  or implied. User hereby grants a royalty-free license to any and all
  derivatives based upon this software code base.


  SNMP++ G A U G E. H

  GAUGE32 CLASS DEFINITION

  DESIGN + AUTHOR:
  Peter E Mellquist

  LANGUAGE:
  ANSI C++

  OPERATING SYSTEMS:
  MS-Windows Win32
  BSD UNIX

  DESCRIPTION:
  Class definition for SMI Gauge32 class.

=====================================================================*/
// $Id: gauge.h,v 1.2 2003/12/01 21:39:23 fock Exp $

#ifndef _GAUGE_H_
#define _GAUGE_H_

#include "integer.h"

//------------[ Gauge32 Class ]------------------------------------------
/**
 * The gauge class allows all the functionality of unsigned integers
 * but is recognized as a distinct SMI type. Gauge32 objects may be
 * set or get into Vb objects.
 */
class DLLOPT Gauge32: public SnmpUInt32
{
 public:

  //-----------[ Constructors and Destrucotr ]----------------------

  /**
   * Constructs a valid Gauge32 with value 0.
   */
  Gauge32() : SnmpUInt32() { smival.syntax = sNMP_SYNTAX_GAUGE32; };

  /**
   * Constructs a valid Gauge32 with the given value.
   *
   * @param ul - value (0..MAX_UINT32)
   */
  Gauge32(const unsigned long ul) : SnmpUInt32(ul)
    { smival.syntax = sNMP_SYNTAX_GAUGE32; };

  /**
   * Copy constructor.
   *
   * @param g32 - value
   */
  Gauge32(const Gauge32 &g32);

  /**
   * Destructor (ensure that SnmpUInt32::~SnmpUInt32() is overridden).
   */
  ~Gauge32() {};

  //-----------[ SnmpSyntax methods ]----------------------

  /**
   * Get the Syntax of the object.
   *
   * @return This method always returns sNMP_SYNTAX_GAUGE32.
   */
  SmiUINT32 get_syntax() const { return sNMP_SYNTAX_GAUGE32; };

  /**
   * Clone the object.
   *
   * @return A cloned Gauge32 object allocated through new.
   */
  SnmpSyntax *clone() const { return (SnmpSyntax *) new Gauge32(*this); };

  //-----------[ Overload some operators ]----------------------

  /**
   * Assign a Gauge32 to a Gauge32.
   */
  Gauge32& operator=(const Gauge32 &uli)
    { smival.value.uNumber = uli.smival.value.uNumber; return *this;};

  /**
   * Assign a unsigned long to a Gauge32.
   *
   * @param ul - New value
   */
  Gauge32& operator=(const unsigned long ul)
    { smival.value.uNumber = ul; return *this; };

  // otherwise, behave like an unsigned int
  /**
   * Cast a Gauge32 to unsigned long.
   *
   * @return Current value as unsigned long.
   */
  operator unsigned long() { return smival.value.uNumber; };

};
#endif // _GAUGE_H_
